# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

from pathlib import Path

import pytest

from parselglossy.documentation import documentation_generator


@pytest.fixture
def template():
    return {
        "keywords": [
            {"name": "title", "type": "str", "docstring": "Title of the calculation."}
        ],
        "sections": [
            {
                "name": "foo",
                "docstring": "brilliant",
                "keywords": [
                    {
                        "name": "tragic",
                        "type": "str",
                        "docstring": "Title of the calculation.",
                        "default": "Wham! Bang! Pow! Let's Rock Out!",
                    }
                ],
                "sections": [
                    {
                        "name": "Bar",
                        "docstring": "A ba-bar section",
                        "keywords": [
                            {
                                "name": "amazing",
                                "type": "str",
                                "docstring": "Title of the calculation.",
                                "default": "Bobson Dugnutt",
                            }
                        ],
                    }
                ],
            }
        ],
    }


def test_documentation(template):
    doc_ref = Path(__file__).parent / Path("ref/simple.rst")
    with doc_ref.open("r") as ref:
        stuff = ref.read().rstrip("\n")

    assert documentation_generator(template) == stuff
